!======================================================================== 
      subroutine ilu0(n,a,colind,rwptr,relax,pivots,diag,ierr)

!======================================================================== 
!                    *** ILU0 ***					
!======================================================================== 
!                                                                     	
!      Incomplete LU factorization for 0 levels of fill-in.	      	
!      This is the version that does not modify the offdiagonal		
!      nonzeros of A, so only the pivot elements need be computed 	
!      and stored.  This corresponds to "-1" levels of fill-in, 	
!      rather than the "0" levels normally referred to.  Note that	
!      a great deal of confusion exists in the literature about what	
!      an ILU(0) preconditioner is .  Here, the preconditioner  	
!      does not allow modification of an off-diagonal entry, even    	
!      it is a nonzero in the original matrix A.                     	
!									
!      For matrices from quasi-uniform regular meshes, the resulting	
!      preconditioner should be the same as ILU(0).			
!                                                                     	
!      The implicit splitting is                                      	
!          LU = (D + LA) (D)^-1 (D + UA)                            	
!             = (I + LA(D)^-1) (D + UA)                            	
!             = (D + LA) (I + (D)^-1 UA)                            	
!       where LA, UA are the strictly lower and 			
!       strictly upper triangular parts of A.                         	
!                                                                     	
!  NOTE: This code will sort the entries in each row of A to be		
!        in column increasing order.	                         	
!  NOTE: A tolerance parameter SMALL is used for handling small       	
!        pivots; this replaces a diagonal entry in some cases.        	
!        SMALL is set to machine tolerance.				
!                                                                     	
!-----------------------------------------------------------------------
!									
!     Randall Bramley						      	
!     Department of Computer Science				      	
!     Indiana University					      	
!     email: bramley@cs.indiana.edu				      	
!     Mon May 30 07:36:14 EST 1994					
!									
!======================================================================== 
!
! parameters
!-----------
!
! on entry:
!========== 
! n       = integer. the dimension of the matrix a.
!
! a,colind,rwptr = matrix stored in compressed sparse row format,
!           each row ordered in increasing column order.
!
! relax   = relaxation parameter.  Gives the multiplier to use before
!           adding discarded off-diagonal modification to the diagonal.
!           Straightforward ILU(-1) uses relax = 1.0d0.
!
! on return:
!=========== 
!
! pivots  = vector of length n of pivot elements
!                                                                        
!  ierr is an error flag:
!        ierr  = -i --> near zero pivot in step i
!        ierr  = 0  --> all's OK
!        ierr  = 1  --> not enough storage; check mneed.
!        ierr  = 2  --> illegal parameter
!
! Work arrays:
!=============
!
!  diag    = Indices of diagonal entries of A in a, colind, rwptr data structure.
!
! Storage requirements:
!======================
!
!   diag:	n 		integers.
!   pivots:     n 		reals.
!
!  Warnings:
!===========
!
! 1) The CSR data structure will be row ordered in increasing column
!    order.  Currently this is done by calls to rowinc, which can mess
!    up any reorderings the user is keeping around.
!
! 2) The diagonal entries of A are assumed to be nonzero., and in particular
!    to have storage provided for them in the CSR data structure.
!
!============================================================================ 
      implicit none

      integer   n, ierr, rwptr(n+1), colind(*), diag(n)
      integer   i, j, k, ie, is
      logical found
      double precision :: pivots(n), A(*), relax, subdiag
      double precision :: SMALL

!     -------------------------------------------------
!     Set divide by zero tolerance to machine tolerance
!     -------------------------------------------------
      SMALL = sqrt(epsilon(SMALL))
      ierr  =  0

!     -------------------------------------------------------
!     Sort the matrix to have rows in increasing column order
!     -------------------------------------------------------
      do k = 1, n
         is = rwptr(k)
         ie = rwptr(k+1)
         call rowinc( ie - is, colind(is), a(is) )
      end do

!     ---------------------------------------------
!     Get the diagonal entry indices in array diag,
!     and get the diagonal entries in array pivots
!     ---------------------------------------------
      do 80 i = 1, n
         do 40 j = rwptr(i), rwptr(i+1)-1
             if (colind(j) .eq. i) then
                 pivots(i) = A(j)
                 diag(i) = j
                 go to 80
             end if 
 40      continue

!         ------------------------------------------------------------
!         If we make it to here, then the diagonal entry of A is not
!         part of the CSR data structure.  This is an error condition.
!         ------------------------------------------------------------
         ierr = -i
         return
 80   continue

!     -------------------------
!     Perform the factorization
!     -------------------------
      do 140 i = 1, n

!        ------------------------------------------------------
!        Invert the i-th pivot element, to avoid extra divides.
!        ------------------------------------------------------
         if (abs(pivots(i)) .le. SMALL) then
            ierr = -i
            pivots(i) = sign(SMALL,pivots(i))
         end if
         pivots(i) = 1.0d0/pivots(i)
!        ----------------------------------------------------------
!        For each nonzero A(i,j) in upper triangular part of row i,
!        update the corresponding diagonal entry A(j,j) by
!            A(j,j) = A(j,j) - A(i,j)*A(j,i)/A(i,i)
!        ----------------------------------------------------------
         do 120 j = diag(i)+1, rwptr(i+1)-1

!        -----------------------------------------------
!        Find the subdiagonal entry A(j,i), it it exists
!        -----------------------------------------------
            found = .false.
            do 100 k = rwptr(colind(j)), diag(colind(j))-1
               if (colind(k) .eq. i) then
                  found = .true.
                  subdiag = A(k)
               end if 
 100        continue
         if (found) pivots(colind(j)) = pivots(colind(j))       &
!                    - subdiag * A(j) * pivots(i)
                    - relax * subdiag * A(j) * pivots(i)
 120     continue
 140  continue

      return

!====================  End of ILU0 ==================================== 

      end
