!======================================================================
      subroutine ilus(n, A, colind, rwptr,        &
                      s, relax,                   &
                      lu, jlu, ilup,              &
                      ierr, mneed)
!**********************************************************************
!                   SPLIB Copyright (C) 1995-1999
!                     Indiana University        
!**********************************************************************

!======================================================================== 
!                                                                  
!      Incomplete LU factorization for s levels of fill-in.	       	
!                                                                 
!======================================================================== 
!                                                                      	
!  This routine performs incomplete LU factorization, with s levels	
!  of fill-in.  It is passed the matrix A in CSR format, and a real*8
!  parameter relax, which will indicate whether or not to perform	
!  MILU factorization.  If the parameter is zero, regular ILU is	
!  performed instead.  A single integer work array is passed in which	
!  to create the LU factors data structures, and for intermediate	
!  storage in the computations.  What is returned are three pointers
!  to the start locations in the work array, for the combined MSR
!  data structure containing the LU factors.  These pointers will be
!  aligned on double word boundaries, provided that the initial word
!  in the work array is aligned on a double word boundary.		
!                                                              
!  If not enough work space is provided, an error code is returned	
!  and the pointers are not set.  An estimate of the additional	
!  work space needed is returned in mneed, but this is only a guess.
!                                                                  
!     Randall Bramley and Xiaoge Wang				       	
!     Department of Computer Science				       
!     Indiana University					       	
!     email: bramley@cs.indiana.edu				   
!     Sun Jan  2 11:09:18 EST 1994				  
!======================================================================== 
!
! parameters
!-----------
!
! on entry:
!========== 
! n       = integer. the dimension of the matrix a.
!
! A,colind,rwptr = matrix stored in compressed sparse row format.
!
! s       = integer; the fill-in parameter.  Fill-in generated during the
!           factorization is kept only if it is caused by a nonzero element
!           in the matrix that was created on the s level or less.  Fill
!           generated by the original nonzeros are of level 1, those created
!           by such entries are level 2, etc.  The original matrix entries
!           are defined as being of level 0.  See symbfac.f for details.
!
! relax   = relaxation parameter.  relax = 0 gives ilu(s), relax = 1
!           gives milu(s), and values between give the multiplier to use
!           before adding discarded fill to the diagonal.
!
! maxstr  = integer giving amount of integer word space available.
!
! intermediate variables:
!========================
!
! rowll   = pointer to temp vector of levels in a row
! lastcol = pointer to temp vector of last column to update a row
! levels  = pointer to temp vector keeping all levels during symbfac.
! colptrs = pointer to temp vector keeping column pointers in numfac.
!
! on return:
!=========== 
!
! lu, jlu = (pointers) matrix stored in modified sparse row (msr) format containing
!           the l and u factors together. the diagonal (stored in
!           lu(1:n) ) is inverted. each i-th row of the lu, jlu matrix 
!           contains the i-th row of l (excluding the diagonal entry=1) 
!           followed by the i-th row of u.  
!                                                                        
! ilup    = (pointer) integer array of length n containing the pointers to        
!           the beginning of each row of u in the matrix lu, jlu. 
!                                                                       
!  ierr is an error flag:
!        ierr  = -i --> near zero pivot in step i
!        ierr  = 0  --> all's OK
!        ierr  = 1  --> not enough storage; check mneed.
!        ierr  = 2  --> illegal parameter
!
! mneed   = contains number of additional storage locations needed for lu
!           in case of failure, or total integer words used from work array
!           in case of success.
!
! Storage requirements:
!======================
!
!   jlu:	nzlu  		integers
!   ilup:	n 		    integers
!   lu:		nzlu 		reals
!
!   temporary arrays
!   ----------------
!   rowll       n               integers
!   lastcol     n               integers
!   levels      nzlu            integers
!   colptrs     n               integers
!
!  Warnings:
!===========
!
! 1) A must have nonzero diagonal entries, at least.  This does not assure
!    completion of the factorization, however.
! 2) The pointers ilup, jlu, lu are given as indices to the work array
!    that is passed in.  So if the integer work array is passed in as
!    work(ptr), you will need to offset those pointers by the operations
!        jlu  = jlu  + ptr -1
!        ilup = ilup + ptr -1
!        lu   = lu   + ptr -1
!    on return.
!
!======================================================================== 

      implicit none
      include '../include/intdbl.inc'

      integer ::    n, s,  ierr, rwptr(*), colind(*)
      double precision ::  A(*), relax
      integer ::   mneed, nzlu, remain, alloc_stat, nnz, estimate
      logical milu

!     ----------------------
!     Pointers in work array
!     ----------------------
      double precision, pointer, dimension(:) ::  lu
      integer, pointer, dimension(:) ::  jlu, ilup
      integer, pointer, dimension(:) ::  lastcol, rowll, levels, colptrs

      ierr  =  0
      nnz = rwptr(n+1)-1

!     ------------------------------------------------
!     If requested level of fill is negative, set to 0
!     ------------------------------------------------
      s = max(s,0)

!     ----------------------------------------------
!     If relaxation parameter is zero, method is ILU
!     ----------------------------------------------
      milu = (relax  /=  0.0d0)

!     ----------------------------------------------------------------
!     Compute pointers into work array.  This version uses two arrays
!     dependent on the (unknown in advance) size nzlu, jlu(1:nzlu) and
!     levels(1:nzlu).  To handle this, first the data space for the
!     work arrays in symbfac is allocated, then half the remainder is
!     assigned to jlu, the rest to levels.  The partitioning of the 
!     integer work array is as:
!
!         [ilup   jlu      ...      levels    rowll lastcol],
!         [-n-----nzlu-----...-------nzlu-------n------n---]
!
!     The index of the last entry in the work array is maxstr, and
!     rowll and lastcol start at the far end of the work array.
!     Note that these are aligned on double word boundaries.
!     ----------------------------------------------------------------
      allocate(lastcol(n), stat=alloc_stat)
      if (alloc_stat /= 0) then
        ierr = 1
        return
      end if
      allocate(rowll(n), stat=alloc_stat)
      if (alloc_stat /= 0) then
        ierr = 1
        return
      end if

!     ----------------------------------------------------------
!     Set ilup to the far left of the avaiable work array space.
!     ----------------------------------------------------------
      allocate(ilup(n), stat=alloc_stat)
      if (alloc_stat /= 0) then
        ierr = 1
        return
      end if

!     -----------------------------------------------------
!     Set up the unknown length arrays jlu, levels
!     mneed = estimate of needed memory.  Exact for ILU(0).
!     -----------------------------------------------------
      estimate =  s*nnz+n+1

 100  continue

      remain = estimate
      allocate(jlu(estimate), stat=alloc_stat)
      if (alloc_stat /= 0) then
        ierr = 1
        return
      end if

      allocate(levels(estimate), stat=alloc_stat)
      if (alloc_stat /= 0) then
        ierr = 1
        return
      end if

!     -------------------------------
!     Perform symbolic factorization:
!     -------------------------------
      call symbfac(n, colind, rwptr, s, remain, nzlu,    &
             jlu, ilup, lastcol, levels, rowll, ierr )

!     -------------------------------------------
!     If not enough storage, try again with more.
!     -------------------------------------------
      if (ierr == 1) then
        deallocate(jlu, levels)
        estimate = 2*estimate
      	go to 100
      end if
      if (ierr /= 0) return

!     -----------------------------------------------------------------
!     Check to see if enough space is available for the lu factorization.
!     The partitioning of the integer work array is as:
!
!        [ilup   jlu           LU        colptrs],
!        [-n-----nzlu-----intdbl*nzlu-------n---]
!
!     -----------------------------------------------------------------
      deallocate(lastcol, rowll, levels)
      allocate(colptrs(n), stat=alloc_stat)
      if (alloc_stat /= 0) then
        ierr = 1
        return
      end if

!     ------------------
!     Set pointer to LU 
!     ------------------
      allocate(lu(nzlu), stat=alloc_stat)
      if (alloc_stat /= 0) then
        ierr = 1
        return
      end if

!     --------------------------------
!     Perform numerical factorization.
!     --------------------------------
      call numfac(n, colind, rwptr, jlu, ilup, A,      &
                    lu, relax, ierr, colptrs, milu)
      deallocate(colptrs)

!     -----------------------------------------------
!     Return actual integer words used in LU factors.
!     -----------------------------------------------
      mneed = (1 + intdbl)*nzlu + n 

      return
      end subroutine ilus
