!=======================================================================
       subroutine ilut (n, a,ja,ia, lfil,tol, alu,jlu,ju, iwk,      &
                        wu,wl,jr,jwl,jwu,ierr)
!-----------------------------------------------------------------------
!                      *** ILUT preconditioner ***          
!                                   
!   Yousef Saad's incomplete LU factorization with dual truncation  
!   mechanism (numerical and positional dropping).  This is modified    
!   only in order to use a different calling sequence, and users    
!   should refer to the SPARSKIT documentation for more information on  
!   the routine.                            
!                                   
!-----------------------------------------------------------------------

!---- coded by Youcef Saad May, 5, 1990. ------------------------------
!---- Dual drop-off strategy works as follows.                         
!                                                                      
!     1) Theresholding in L and U as set by tol. Any element whose size
!        is less than some tolerance (relative to the norm of current  
!        row in u) is dropped.                                         
!                                                                      
!     2) Keeping only the largest lfil+il(i) elements in the i-th row  
!        of L and the largest lfil+iu(i) elements in the i-th row of   
!        U where il(i), iu(i) are the original number of nonzero       
!        elements of the L-part and the U-part of the i-th row of A    
!                                                                      
! Flexibility: one can use tol=0 to get a strategy based on keeping the
! largest elements in each row of L and U. Taking tol .ne. 0 but lfil=n
! will give the usual threshold strategy (however, fill-in is then     
! impredictible).                                                      
!                                                                      
!----------------------------------------------------------------------
! PARAMETERS
!-----------
!
! on entry:
!==========
! n       = integer. The dimension of the matrix A.
!
! a,ja,ia = matrix stored in Compressed Sparse Row format.
!
! lfil    = integer. The fill-in parameter. Each row of L and
!           each row of U will have a maximum of lfil elements
!           in addition to their original number of nonzero elements.
!           Thus storage can be determined beforehand.
!           lfil must be .ge. 0.
!
! iwk     = integer. The minimum length of arrays alu and jlu
!
! On return:
!===========
!
! alu,jlu = matrix stored in Modified Sparse Row (MSR) format containing
!           the L and U factors together. The diagonal (stored in
!           alu(1:n) ) is inverted. Each i-th row of the alu,jlu matrix
!           contains the i-th row of L (excluding the diagonal entry=1)
!           followed by the i-th row of U.
!
! ju      = integer array of length n containing the pointers to
!           the beginning of each row of U in the matrix alu,jlu.
!
! ierr    = integer. Error message with the following meaning.
!           ierr  = 0    --> successful return.
!           ierr .gt. 0  --> zero pivot encountered at step number ierr.
!           ierr  = -1   --> Error. input matrix may be wrong.
!                            (The elimination process has generated a
!                            row in L or U whose length is .gt.  n.)
!           ierr  = -2   --> The matrix L overflows the array al.
!           ierr  = -3   --> The matrix U overflows the array alu.
!           ierr  = -4   --> Illegal value for lfil.
!           ierr  = -5   --> zero row encountered.
!
! work arrays:
!=============
! jr,jwu,jwl      = integer work arrays of length n.
! wu, wl          = real work arrays of length n+1, and n resp.
!
! Notes:
! ------
! A must have all nonzero diagonal elements.
!-----------------------------------------------------------------------

        implicit none
        integer n
        double precision :: a(*), alu(*), wu(n+1), wl(n), tol
        integer ja(*),ia(n+1),jlu(*),ju(n),jr(n), jwu(n),       &
            jwl(n), lfil, iwk, ierr
        integer lenu, lenl, jrow, nl
        integer jj, j1, j2, len, ju0, j, ii, k, lenl0, lenu0, jpos
        double precision :: s, fact, t, tnorm

        if (lfil .lt. 0) goto 998
!-------------------------------
! initialize ju0 (points to next element to be added to alu,jlu)
! and pointer.
!
    ju0 = n+2
    jlu(1) = ju0
!
!  integer double pointer array.
!
    jr(1:n) = 0
!-----------------------------------------------------------------------
!  beginning of main loop.
!-----------------------------------------------------------------------
    do 500 ii = 1, n
           j1 = ia(ii)
           j2 = ia(ii+1) - 1
           tnorm = 0.0d0
           do 501 k=j1,j2
              tnorm = tnorm+abs(a(k))
 501          continue
              if (tnorm .eq. 0.0) goto 999
              tnorm = tnorm/real(j2-j1+1)
!
!--- unpack L-part and U-part of row of A in arrays wl, wu --
!
       lenu = 1
       lenl = 0
       jwu(1) = ii
       wu(1) = 0.0
       jr(ii) = 1
!
       do 170  j = j1, j2
           k = ja(j)
           t = a(j)
           if (abs(t) .lt. tol*tnorm .and. k .ne. ii) goto 170
           if (k .lt. ii) then
              lenl = lenl+1
              jwl(lenl) = k
              wl(lenl) = t
              jr(k) = lenl
           else if (k .eq. ii) then
              wu(1) = t
           else
              lenu = lenu+1
              jwu(lenu) = k
              wu(lenu) = t
              jr(k) = lenu
           endif
 170    continue
        tnorm = tnorm/real(j2-j1+1)
        lenl0 = lenl
        lenu0 = lenu
        jj = 0
        nl = 0
!-------------------------------------------------------------------
!---------------------- eliminate previous rows --------------------
!-------------------------------------------------------------------
 150    jj = jj+1
        if (jj .gt. lenl) goto 160
!-------------------------------------------------------------------
! in order to do the elimination in the correct order we need to
! exchange the current row number with the one that has
! smallest column number, among jj,jj+1,...,lenl.
!-------------------------------------------------------------------
        jrow = jwl(jj)
        k = jj
!
! determine smallest column index
!
        do 151 j=jj+1,lenl
           if (jwl(j) .lt. jrow) then
              jrow = jwl(j)
              k = j
           endif
 151    continue
!
! exchange in jwl
!
       if (k .ne. jj) then
           j = jwl(jj)
           jwl(jj) = jwl(k)
           jwl(k) = j
!
! exchange in jr
!
           jr(jrow) = jj
           jr(j) = k
!
! exchange in wl
!
           s = wl(jj)
           wl(jj) = wl(k)
           wl(k) = s
        endif
!
        if (jrow .ge. ii) goto 160
!---------get the multiplier for row to be eliminated: jrow
        fact = wl(jj)*alu(jrow)
! zero out element in row by setting jr(jrow) = 0
        jr(jrow) = 0
        if (abs(fact)*wu(n+2-jrow) .le. tol*tnorm) goto 150
!-------------------------------------------------------------------
!------------ combine current row and row jrow ---------------------
!-------------------------------------------------------------------
        do 203 k = ju(jrow), jlu(jrow+1)-1
           s = fact*alu(k)
           j = jlu(k)
           jpos = jr(j)
!
! if fill-in element is small then disregard:
!
           if (abs(s) .lt. tol*tnorm .and. jpos .eq. 0) goto 203
           if (j .ge. ii) then
!
!     dealing with upper part.
!
              if (jpos .eq. 0) then
!     this is a fill-in element
                 lenu = lenu+1
                 if (lenu .gt. n) goto 995
                 jwu(lenu) = j
                 jr(j) = lenu
                 wu(lenu) = - s
              else
!     no fill-in element --
                 wu(jpos) = wu(jpos) - s
              endif
           else
!
!     dealing with lower part.
!
              if (jpos .eq. 0) then
!     this is a fill-in element
                 lenl = lenl+1
                 if (lenl .gt. n) goto 995
                 jwl(lenl) = j
                 jr(j) = lenl
                 wl(lenl) = - s
              else
!     no fill-in element --
                 wl(jpos) = wl(jpos) - s
              endif
           endif
 203    continue
        nl = nl+1
        wl(nl) = fact
        jwl(nl)  = jrow
    goto 150
!----------------------------------------------------------
!------------ update l-matrix -----------------------------
!----------------------------------------------------------
 160    len = min0(nl,lenl0+lfil)
! 160    len = min0(nl,lfil)

    call qsplit (wl,jwl,nl,len)
!
        do 204 k=1, len
           if (ju0 .gt. iwk) goto 996
           alu(ju0) =  wl(k)
           jlu(ju0) =  jwl(k)
           ju0 = ju0+1
 204    continue
!
!  save pointer to beginning of row ii of U
!
        ju(ii) = ju0
!
!  reset double-pointer jr to zero (L-part - except first
!  jj-1 elements which have already been reset)
    do 306 k= jj, lenl
              jr(jwl(k)) = 0
 306    continue
        len = min0(lenu,lenu0+lfil+1) ! +1 added 11/23/94
!        len = min0(lenu,lfil)
    call qsplit (wu(2), jwu(2), lenu-1,len)
!----------------------------------------------------------
!------------ update u-matrix -----------------------------
!----------------------------------------------------------
        t = abs(wu(1))
        if (len + ju0 .gt. iwk) goto 997
        do 302 k=2, len
           jlu(ju0) = jwu(k)
           alu(ju0) = wu(k)
           t = t + abs(wu(k) )
           ju0 = ju0+1
 302    continue
!
!     save norm in wu (backwards). Norm is in fact average abs value
!
        wu(n+2-ii) = t / real(len+1)
!
!     store inverse of diagonal element of u
!
        if (wu(1) .eq. 0.0) wu(1) = (0.0001 + tol)*tnorm
!
        alu(ii) = 1.0d0/ wu(1)
!
!     update pointer to beginning of next row of U.
!
    jlu(ii+1) = ju0
!
!     reset double-pointer jr to zero (U-part)
!
    do 308 k=1, lenu
           jr(jwu(k)) = 0
 308    continue
!-----------------------------------------------------------------------
!     end main loop
!-----------------------------------------------------------------------
 500  continue
        
!
! change by tjl -- set iwk to the the number of entries actually used.
!
        iwk = ju0 - 1
        ierr = 0
        return
!
!     zero pivot :
!
! 900    ierr = ii
!        return
!
!     incomprehensible error. Matrix must be wrong.
!
 995    ierr = -1
        return
!
!     insufficient storage in L.
!
 996    ierr = -2
        return
!
!     insufficient storage in U.
!
 997    ierr = -3
        return
!
!     illegal lfil entered.
!
 998    ierr = -4
        return
!
!     zero row encountered
!
 999    ierr = -5
        return
!---------------- end of ilut  -----------------------------------------
!-----------------------------------------------------------------------
        end
!----------------------------------------------------------------------

        subroutine qsplit  (a, ind, n, ncut)
        double precision :: a(n)
        integer ind(n), n, ncut
!-----------------------------------------------------------------------
!     does a quick-sort split of a real array.
!     on input a(1:n). is a real array
!     on output a(1:n) is permuted such that its elements satisfy:
!
!     abs(a(i)) .ge. abs(a(ncut)) for i .lt. ncut and
!     abs(a(i)) .le. abs(a(ncut)) for i .gt. ncut
!
!     ind(1:n) is an integer array which permuted in the same way as a(*).
!-----------------------------------------------------------------------
        double precision :: tmp, abskey
        integer itmp, first, last
!-----
        first = 1
        last = n
        if (ncut .lt. first .or. ncut .gt. last) return
!
!     outer loop -- while mid .ne. ncut do
!
 1      mid = first
        abskey = abs(a(mid))
        do 2 j=first+1, last
           if (abs(a(j)) .gt. abskey) then
              mid = mid+1
!     interchange
              tmp = a(mid)
              itmp = ind(mid)
              a(mid) = a(j)
              ind(mid) = ind(j)
              a(j)  = tmp
              ind(j) = itmp
           endif
 2      continue
!
!     interchange
!
        tmp = a(mid)
        a(mid) = a(first)
        a(first)  = tmp
!
        itmp = ind(mid)
        ind(mid) = ind(first)
        ind(first) = itmp
!
!     test for while loop
!
        if (mid .eq. ncut) return
        if (mid .gt. ncut) then
           last = mid-1
        else
           first = mid+1
        endif
        goto 1
!----------------end-of-qsplit------------------------------------------
!-----------------------------------------------------------------------
        end
