!========================================================================
      subroutine tridiag(n,A,colind,rwptr,quad,ipvt,iskip,ierr)
!========================================================================

!*********************************************************************
!                   SPLIB Copyright (C) 1995-1999                    *
!                     Indiana University                             *
!*********************************************************************

!                 *** Tridiagonal preconditioner ***            
!----------------------------------------------------------------------== 
!                                   
!  This implements block diagonal preconditioning, where        
!  each diagonal block is from the corresponding tridiagonal part   
!  of the matrix A.  The size of the blocks is given implicitly in  
!  the integer iskip, stored in jlu(n).  Factors are set up for     
!  contiguous blocks of iskip rows and so by using iskip = n the    
!  factor can be for the entire matrix's tridiagonal part.      
!  Using iskip = 1 is an inefficient way of performing diagonal     
!  preconditioning, while 1 < iskip < n is for a block tridiagonal  
!  preconditioner.  Gaussian elimination with pivoting is used for  
!  factoring each block, so it is necessary to store the pivot      
!  sequence in jlu(1:n-1).  quad has four vectors giving the nonzeros   
!  of the tridiagonal preconditioner.  The method performs Gaussian 
!  elimination with partial pivoting on the tridiagonal part of the 
!  matrix; this is used as a preconditioner via routines tridsl.f and   
!  trifac.f.                                
!                                   
!  This is generally not a very good preconditioner!            
!                                   
!----------------------------------------------------------------------==
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Tue Jan 18 11:14:46 EST 1994
!
!----------------------------------------------------------------------==
!
! parameters
!-----------
!
! on entry:
!========== 
! n       = integer. the dimension of the matrix a.
!
! A,colind,rwptr = matrix stored in compressed sparse row format.
!
! iskip   = integer specifying the block size to use for a block
!           diagonal preconditioner, with each block a tridiagonal matrix.
!
! on return:
!=========== 
!
! quad    = matrix storing the four diagonal needed for the LU factors
!           of the tridiagonal part of the matrix A.  quad must be n x 4
!           double precision.
!
! ierr is an error flag:
!        ierr  = -i --> near zero pivot in step i
!        ierr  = 0  --> all's OK
!        ierr  = 1  --> not enough storage; check mneed.
!        ierr  = 2  --> illegal parameter
!
! work arrays:
!=============
!
! ipvt     = integer work array of length >= n.
!            The integer iskip is placed in ipvt(n), for passing to the
!            preconditioner application programs.
!
!----------------------------------------------------------------------- 

      implicit none
      integer   n, ipvt(n), ierr, rwptr(n+1), colind(*),k,kk
      double precision :: A(*), quad(n,*)
      integer   length, iskip, job, ndz, ioff, blksz

!=======================================================================
!     Start of Executable Statements
!=======================================================================

      ierr  =  0
      if (iskip .le. 0) ierr = 2
      if (n .le. 0) ierr = 2
      if (ierr .ne. 0) return
      blksz = min(n,iskip)

!     -----------------------------------------------------------------
!     Extract the central tridiagonal part of A:
!     Mapping to quad: quad(*,1) --> first subdiagonal
!                      quad(*,2) --> main diagonal
!                      quad(*,3) --> first superdiagonal
!                      quad(*,4) --> second superdiagonal (for fill-in)
!     -----------------------------------------------------------------
      job = 0
      ioff = -1
      call getdia(n,n,job,A,colind,rwptr,ndz,quad(1,1),ipvt,ioff)
      ioff = 0
      call getdia(n,n,job,A,colind,rwptr,ndz,quad(1,2),ipvt,ioff)
      ioff = 1
      call getdia(n,n,job,A,colind,rwptr,ndz,quad(1,3),ipvt,ioff)

!     ----------------------------------------------------------------
!     Perform the LU factorization.  Note that because getdia extracts
!     the diagonal assuming they start at row number 1,
!     the first subdiagonal     begins in quad(2,1)
!     the main diagonal         begins in quad(1,2)
!     the first superdiagonal   begins in quad(1,3)
!     ----------------------------------------------------------------

      do 100 k = 1, n, blksz

!        ------------------------------------------
!        Length of the tridiagonal matrix to factor
!        ------------------------------------------
         length = min(blksz,n-k+1)

!        --------------------------------------------------
!        Offset into the quad array for this block to start
!        --------------------------------------------------
         ioff = k

!        --------------------
!        Do the factorization
!        --------------------
         call trifac(length,quad(1+ioff,1),     &
                            quad(ioff,2),       &
                            quad(ioff,3),       &
                            quad(ioff,4),       &
                            ipvt(ioff),ierr)

!        ----------------
!        Check error code
!        ----------------
         if (ierr.ne. 0) then
            write(6,*) 'Error condition in trifac on block beginning'
            write(6,*) '       at position ',k
            write(6,*) '       Occurs in tridiag.f'
            write(6,*) '       Error code = ',ierr
            write(6,*) '       Setting block to identity'
            do kk = ioff,ioff+length-1
               quad(kk,1) = 0.0d0
               quad(kk,2) = 1.0d0
               quad(kk,3) = 0.0d0
               quad(kk,4) = 0.0d0
               ipvt(kk) = kk-ioff+1
            end do
            ierr = 0
         end if
 100     continue
         ipvt(n) = blksz


 1000  format(128(2(4(d16.8,3x),/),/))
 1010  format(128(8(i5,3x),/))
      return
      end
!
!
!========================================================================
!
!
      subroutine trifac( n, dl, d, du, du2, ipiv, info )
!========================================================================
!       **** Factor a tridiagonal matrix, with pivoting ***     *
!========================================================================
!
!     Factor a linear system Ax = b, where A is tridiagonal and has
!     been factored by trifac.  This routine is modified from a Lapack
!     routine (dgttrf.f).  The factorization uses partial pivoting, so
!     an extra vector of storage is needed.  Furthermore, the factors
!     overwrite the input matrix.  This routine is from LAPACK, with
!     some minor changes.
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Tue Jan 18 10:01:23 EST 1994
!
!  -- original lapack routine (version 1.1) --
!     univ. of tennessee, univ. of california berkeley, nag ltd.,
!     courant institute, argonne national lab, and rice university
!     march 31, 1993
!
!========================================================================
!
!
!  Scalar arguments
      implicit none
      integer            info, n
!  Array arguments 
      integer            ipiv( * )
      double precision   d( * ), dl( * ), du( * ), du2( * )
!
!  purpose
!  =======
!
!  trifac computes an lu factorization of a real tridiagonal matrix a
!  using elimination with partial pivoting and row interchanges.
!
!  the factorization has the form
!     a = l * u
!  where l is a product of permutation and unit lower bidiagonal
!  matrices and u is upper triangular with nonzeros in only the main
!  diagonal and first two superdiagonals.
!
!  arguments
!  =========
!
!  n       (input) integer
!          the order of the matrix a.
!
!  dl      (input/output) double precision array, dimension (n-1)
!          on entry, dl must contain the (n-1) subdiagonal elements of
!          a.
!          on exit, dl is overwritten by the (n-1) multipliers that
!          define the matrix l from the lu factorization of a.
!
!  d       (input/output) double precision array, dimension (n)
!          on entry, d must contain the diagonal elements of a.
!          on exit, d is overwritten by the n diagonal elements of the
!          upper triangular matrix u from the lu factorization of a.
!
!  du      (input/output) double precision array, dimension (n-1)
!          on entry, du must contain the (n-1) superdiagonal elements
!          of a.
!          on exit, du is overwritten by the (n-1) elements of the first
!          superdiagonal of u.
!
!  du2     (output) double precision array, dimension (n-2)
!          on exit, du2 is overwritten by the (n-2) elements of the
!          second superdiagonal of u.
!
!  ipiv    (output) integer array, dimension (n)
!          the pivot indices; for 1 <= i <= n, row i of the matrix was
!          interchanged with row ipiv(i).  ipiv(i) will always be either
!          i or i+1; ipiv(i) = i indicates a row interchange was not
!          required.
!
!  info    (output) integer
!          = 0:  successful exit
!          < 0:  if info = -i, the i-th argument had an illegal value
!          > 0:  if info = i, u(i,i) is exactly zero. the factorization
!                has been completed, but the factor u is exactly
!                singular, and division by zero will occur if it is used
!                to solve a system of equations.
!
!========================================================================
!========================================================================
!
!  Local scalars:
      integer   i
      double precision   fact, temp
!  Intrinsic functions:
      intrinsic abs
!  Parameters:
      double precision   zero
      parameter ( zero = 0.0d+0 )
!
!========================================================================
!
      info = 0
      if( n.lt.0 ) then
         info = -1
         write(6,*) 'Error in trifac; info = ',info
         return
      end if
!
!     quick return if possible
!
      if( n.eq.0 ) return
!
!     initialize ipiv(i) = i
!
      do 10 i = 1, n
         ipiv( i ) = i
   10 continue
!
      do 20 i = 1, n - 1
         if( dl( i ).eq.zero ) then
!
!           subdiagonal is zero, no elimination is required.
!
            if( d( i ).eq.zero .and. info.eq.0 ) info = i
!            if( i.lt.n-1 ) du2( i ) = zero
         else if( abs( d( i ) ).ge.abs( dl( i ) ) ) then
!
!           no row interchange required, eliminate dl(i)
!
            fact = dl( i ) / d( i )
            dl( i ) = fact
            d( i+1 ) = d( i+1 ) - fact*du( i )
!            if( i.lt.n-1 ) du2( i ) = zero
         else
!
!           interchange rows i and i+1, eliminate dl(i)
!
            fact = d( i ) / dl( i )
            d( i ) = dl( i )
            dl( i ) = fact
            temp = du( i )
            du( i ) = d( i+1 )
            d( i+1 ) = temp - fact*d( i+1 )
            if( i.lt.n-1 ) then
               du2( i ) = du( i+1 )
               du( i+1 ) = -fact*du( i+1 )
            end if
            ipiv( i ) = ipiv( i ) + 1
         end if
   20 continue
!
      if( d( n ).eq.zero .and. info.eq.0 ) then
         info = n
         return
      end if
!
      return
!
!  ---------------> End of trifac <-----------------
!
      end
