!=======================================================================
      subroutine utltsl (n, y, x, lu, jlu, uptr, pcmeth, a, colind, rwptr)

!**********************************************************************
!                   SPLIB Copyright (C) 1995-1999                    **
!                     Indiana University                             **
!**********************************************************************

!=======================================================================*
!                *** UTLTSL: Solve M'x = y ***              *
!                                   *
!  Apply transpose preconditioner, for methods such as CGNE which   *
!  require transpose matrix operations.                 *
!                                   *
!=======================================================================*
!
!  x       : double precision :: vector of length n (output).
!  y       : double precision :: vector of length n (input).
!
!  n       : Order of matrix A, and system Mx = y where M is the
!            preconditioning matrix.
!  A       : double precision :: array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!  uptr, 
!  lu,jlu  : Data structure for the preconditioner M, with different meanings
!            for different preconditioners.
!
!            ILU(s), MILU(s), or ILUt --->
!                 (lu,jlu) gives the combined L/U factors for the preconditioner
!                 in a modified sparse row (MSR) format.  L is unit lower 
!                 triangular and U is upper triangular.  The diagonal of U is
!                 stored in lu(1:n) inverted, to replace divisions in the algorithm
!                 with multiplications.  Each i-th row of the combined L/U matrix in
!                 the (lu,jlu) data structure contains the i-th row of L (excluding
!                 the diagonal entry = 1), followed by the i-th row of u.
!                 uptr is a integer vector of length n containing pointers to the
!                 start position of each row of U in (lu, jlu).
!
!            SSOR(omega)   --->
!                 lu is the scalar acceleration parameter omega.
!                 jlu is not used, and uptr is a vector of pointers to the
!                 diagonal entries in the CSR data structure for A.
!
!            TRID(s)   --->
!                 This implements block diagonal preconditioning, where
!                 each diagonal block is from the corresponding tridiagonal part
!                 of the matrix A.  The size of the blocks is given implicitly in
!                 the integer iskip, stored in jlu(n).  Factors are set up for 
!                 contiguous blocks of iskip rows and so by using iskip = n the 
!                 factor can be for the entire matrix's tridiagonal part.
!                 Using iskip = 1 is an inefficient way of performing diagonal
!                 preconditioning, while 1 < iskip < n is for a block tridiagonal
!                 preconditioner.  Gaussian elimination with pivoting is used for 
!                 factoring each block, so it is necessary to store the pivot
!                 sequence in jlu(1:n-1).  lu has the four vectors giving the nonzeros
!                 of the tridiagonal preconditioner.  Vector uptr is not used.
!
!            ILU0   --->
!                 This implements the version of ILU(0) that does not change any
!                 off-diagonal entries of A, and so only needs to store the n-vector
!                 of diagonal pivots.  lu(1:n) stores those pivots, and jlu(1:n) contains
!                 pointers to the diagonal entry in each row of A.
!
!  pcmeth  : preconditioning method to use.
!       pcmeth = 0  ---> no preconditioner, i.e. m = identity matrix.
!       pcmeth = 1  ---> ILU(c).
!       pcmeth = 2  ---> MILU(c).
!       pcmeth = 3  ---> ILUT(c).
!       pcmeth = 4  ---> SSOR(omega).
!       pcmeth = 5  ---> TRID.
!       pcmeth = 6  ---> ILU0.
!       pcmeth = 7  ---> ECIMGS.
!
!  Routines
!  ========
!
!  lutsolt : Solves u^t l^t x = y.
!            used with ILU, MILU, ILUT, ECIMGS
!  lutssor : Solves u^t l^t x = y.
!            used with SSOR
!  lutsol0 : Solves u^t l^t x = y.
!            used with ILU0
!  tridsl  : Applies transposed tridiagonal block preconditioner.
!  dcopy   : Blas routine for copying vectors
!
!-------------------------------------------------------------------------
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University, Bloomington
!     bramley@cs.indiana.edu
!     Wed Jul 12 10:03:18 EST 1995
!
!-------------------------------------------------------------------------

!     Declarations of arguments
!     -------------------------
      implicit none
      integer n
      double precision :: x(n),y(n),lu(*)
      double precision :: a(*)
      integer jlu(*),uptr(n)
      integer colind(*), rwptr(n+1)
      integer pcmeth, ierr

!     Local variables
!     ---------------
      integer k, length

!     Timing variables
!     ----------------
      include '../include/heads.inc'
      double precision :: mytime, tstart
      external mytime


!========================================================================
!      Beginning of Executable Statements
!========================================================================

!      tstart = mytime()

!     ----------------
!     Copy y over to x
!     ----------------
      call dcopy(n,y,1,x,1)

!     ------------------------------------------
!     Unknown pcmeth (should not be encountered)
!     ------------------------------------------
      if (pcmeth.lt.0 .or. pcmeth .gt. 7) then
         write(6,*) 'Unknown method in utltsl'
         stop
      endif

!     ------------------
!     No preconditioning
!     ------------------
      if (pcmeth.eq.0) then
         return
      endif

!     --------------------------
!     ILU, MILU, ILUT, or ECIMGS
!     --------------------------
      if ( ((pcmeth.ge.1) .and. (pcmeth.le.3)) .or. pcmeth .eq. 7)      &
         call lutsolt (n, x, lu, jlu, uptr)

!     --------------------
!     SSOR preconditioning
!     --------------------
      if (pcmeth.eq.4) call lutssor(n, x, lu, jlu, a, colind, rwptr)

!     --------------------
!     ILU0 preconditioning
!     --------------------
      if (pcmeth.eq.6) call lutsol0(n, x, lu, jlu, a, colind, rwptr)

!     --------------------
!     TRID preconditioning
!     --------------------
      if (pcmeth.eq.5) then
         do 100 k = 1, n, jlu(n)
            length = min(jlu(n),n-k+1)
            call tridsl('T',length,lu(1+k),         &
                                   lu(k+n),         &
                                   lu(k+2*n),       &
                                   lu(k+3*n),       &
                                   jlu(k),x(k),ierr)
            if (ierr .ne. 0) then
               write(6,*)'Error in preconditioner; tridsl ierr = ' ,ierr
               write(6,*)'        length = ', length
               stop
            end if
 100     continue
      end if

!      times(5) = times(5) + mytime() - tstart

      return

!===================== End of LUSL =====================================

      end

!====================================================================== 
      subroutine lutsolt (n, x, lu, jlu, uptr) 
!====================================================================== 
!
!   Perform u^t l^t x = x for LU matrix as produced by ILU, MILU, ILUT,
!   or ECIMGS.
!
!====================================================================== 

        implicit none
    integer n, jlu(*),uptr(n),i,k 
        double precision :: x(n), lu(*)

!       -------------
!       Solve U'x = x
!       -------------
    do i = 1, n
           x(i) = lu(i)*x(i)
       do k=uptr(i),jlu(i+1)-1 
              x(jlu(k)) = x(jlu(k)) - lu(k)*x(i)
           end do
        end do

!       -------------
!       Solve L'x = x
!       -------------
        do i = n,1,-1 
           do k = jlu(i), uptr(i)-1 
              x(jlu(k)) = x(jlu(k)) - lu(k)* x(i) 
           end do
        end do

    return

!================ End of lutsolt =========================================

    end


!======================================================================= 
      subroutine lutsol0(n, x, pivots, diag, a,colind,rwptr) 
!======================================================================= 
!  Solve the system (LU)'x = x, where L and U were created using
!  ILU(-1) preconditioner, the space saver version of ILU(0).
!  The implicit splitting is
!          LU = (I + LA(D)^-1) (D + UA)
!  where D is the diagonal matrix of pivots, and LA, UA are the strictly
!  lower and upper triangular parts of A.
!======================================================================= 

      implicit none
      integer n, colind(*), rwptr(n+1), diag(n)
      double precision :: x(n), pivots(n)
      double precision :: a(*), t
      integer i, j

      do 40 i = 1, n
         t = pivots(i)*x(i)
         do 20 j = diag(i)+1,rwptr(i+1)-1
            x(colind(j)) = x(colind(j)) - t*a(j)
  20     continue
  40  continue

      do 80 i = n, 1, -1
         x(i) = pivots(i)*x(i)
         do 60 j = rwptr(i), diag(i)-1
            x(colind(j)) = x(colind(j)) - a(j)*x(i)
  60     continue
  80  continue
      return

!=================== End of lutsol0 ======================================= 

      end


!======================================================================= 
      subroutine lutssor(n, x, omega, diag, a, colind, rwptr)
!======================================================================= 
!
!  Solve the transposed SSOR preconditioning system
!   M'x = x, where M' =  (I + U' * omega * D^-1) * ( D/omega + L')
!
!======================================================================= 
      implicit none
      integer n, colind(*), rwptr(n+1), diag(n)
      double precision :: x(n), omega
      double precision :: a(*), dk
      integer k, j

      do 40 k = 1, n
         dk = x(k)*omega/a(diag(k))
         do 20 j = diag(k)+1, rwptr(k+1)-1
           x(colind(j)) = x(colind(j)) - dk*a(j)
  20     continue
  40  continue

      do 80 k = n, 1, -1
         dk = x(k)*omega/a(diag(k))
         x(k) = dk
         do 60 j = rwptr(k), diag(k)-1
           x(colind(j)) = x(colind(j)) - dk*a(j)
  60     continue
  80  continue
      
      return
      end
!=================== End of lutssor ======================================= 
