!=========================================================================
      subroutine  cgs (A,colind,rwptr,n,                    &
                       x,r,b,                               &
                       tolrnc,reltv,maxits,ierr,            &
                       work,wksz,                           &
                       lu,jlu,uptr,pcmeth, iunit,instlvl)

!-----------------------------------------------------------------------
!                             *** CGS ***                               
!  The conjugate gradient squared algorithm of Sonneveld and van der    
!  Vorst.                               
!                                   
!  This version differs from the one given in Templates, because it 
!  uses a different change of variables in applying preconditioning.    
!  The effect is that the residual vector r that is used here is the    
!  preconditioned residual, while in Templates it is the        
!  unpreconditioned residual.  Mathematically the two are equivalent,   
!  but in practice this implementation seems to provide better      
!  convergence.  The advantage of the Templates version is that an  
!  estimate of the actual residual is constantly being kept.        
!                                   
!                                   
!-----------------------------------------------------------------------
!
!  Arguments:
!  ----------
!
!     (Data structure for A)
!  n       : Order of matrix A.
!  A       : double precision :: array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!
!     (Required vectors)
!  x       : Vector of unknowns to be solved for.  Contains initial
!            estimate of x on entry, and contains final approximate
!            on return.
!  r       : Residual vector r = b - A*x.
!  b       : Right hand side vector of system to be solved: Ax = b
!  work    : Double precision array of additional storage.  Bicg
!            requires 6 vectors of length n, and if instrumentation
!            routine instr.f is called, an additional 2 vectors of
!            length n are required.  If not enough storage is given,
!            nothing is done and an error code is returned specifying
!            the additional storage needed (in double precision words).
!  wksz    : Length of array work, in double precision words.  
!
!     (Preconditioner data structure)
!  lu,jlu  : Preconditioner in Modified Sparse Row format containing
!            the L and U factors together. The diagonal of U (stored in
!            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
!            contains the i-th row of L (excluding the diagonal entry) 
!            followed by the i-th row of U.  The diagonal of L has
!            all ones, and so is not stored explicitly.
!  uptr    : Integer array of length n containing the pointers to        
!            the beginning of each row of U in the matrix lu,jlu. 
!  pcmeth  : Preconditioning method to use.  See splib.f documentation
!            for the interpretation of this integer.
!
!     (Iteration control)
!  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
!            passed in as a vector in case future versions
!            set pdr and SMALL.  Since Krylov methods generally
!            use the dotproduct r'*r, a local variable tolsq is used
!            in order to avoid square roots.  Note: because the
!            preconditioned residual norm || inv(M)*r || is the natural
!            one tracked by the iterative method, but users generally
!            want the true residual norm || r || as the stopping test,
!            this routine will exit the iterations when the precond
!            residual is small enough and then test the true residual.
!            If it is not sufficiently reduced, then the local tolerance
!            tolsq is reduced and iterations resume.  This is done
!            a maximum of eight times before failure is declared and
!            the routine returns.
!  reltv   : Logical variable indicating whether or not to use a
!            relative residual as the stopping test.  If this is
!            false, the stopping test is || r_k || < tolrnc(1).  If it
!            is true, the test is || r_k || < || r_0 || * tolrnc(1).
!            Note that it is possible to modify SPLIB to use other
!            stopping tests (see routine instr.f).
!  maxits  : Maximum number of iterations allowed.
!
!      (Other stuff)
!  ierr    : Error flag; see below.
!  iunit   : Integer vector of unit numbers for the instrumentation
!            routine instr.f to use.  Not referenced by cgs, just
!            passed along to instr.f
!  instlvl : Integer giving instrumentation level.  If instlvl = 0,
!            then instr.f is not called and array work() need not
!            provide the 2n double precision words used by instr.f
!
!  Some internal variables:
!  ------------------------
!
!  early   : Logical parameter that indicates if a tentative exit
!            was made because of early termination.  This is needed
!            because of the way SPLIB uses unpreconditioned residuals
!            as the stopping test.
!  first   : Logical parameter that specifies to instr.f if this is
!            the first iteration, and so if new headers need to be
!            written to the Xgraph output files.
!  SMALL   : Parameter that determines when it is not safe to divide
!            by a scalar quantity.  Error condition is set if a divisor
!            is smaller than SMALL.
!  pdr     : Parameter to determine if preconditioner is no good.
!            Experience shows that if the preconditioned residual norm
!            is more than 1/eps, where eps is the machine epsilon, then
!            even if the preconditioned residual converges, the original
!            system residual does not.  If you are not comfortable with
!            this try setting pdr = 1.0d200.  
!  urnorm  : Unpreconditioned initial residual.  Used for tests to see
!            if initial residual is good enough, and if preconditioner
!            is no good (see pdr above).
!
!  rh, q,
!  ph, t,
!  v, p,
!  uh, u   : Integer pointers in the array work(*) that give the
!            starting position of the corresponding vector used by CGS.
!  wins1,
!  wins2   : Integer pointers in the array work(*) that give starting
!            positions of two vectors used by instrumentation routines.
!
!======================================================================
!
! Return values:                                                           
! --------------
!
!  x      : Contains an approximate solution (upon successful return).  
!
!  ierr   : Error flag with the following meanings:          
!           ierr = -m --> Not enough workspace provided; m more double
!                         precision words needed in array work().
!           ierr = 0  --> Successful return; tolerance achieved.
!           ierr = 1  --> The initial guess satisfies the stopping test.
!                         This is often a clue that rhs not set properly.
!           ierr = 2  --> Tolerance reset 8 times in attempt to
!                         get true residual small enough.
!           ierr = 3  --> Convergence not achieved in maxits iterations.  
!           ierr = 4  --> Preconditioned residual is 1/eps times
!                         larger than unpreconditioned residual.
!           ierr = 5  --> Convergence not achieved due to tau = 0
!           ierr = 6  --> Convergence not achieved due to rho = 0
!
!  maxits : Contains the number of actual iterations performed.
!  tolrnc : May be modified if reltv = .true.
!                                                                       
!======================================================================
!  Routines
!  --------
!
!  lusl    : Solves M z = r.
!  bmux    : Computes matrix-vector product with A.
!  instr   : Instrumentation routine for timing, etc.
!  BLAS routines: ddot, daxpy, dcopy
!
!======================================================================
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Mon Jun  6 11:01:59 EST 1994
!     Documented Sun Jul  9 15:48:33 EST 1995
!
!----------------------------------------------------------------------
!
!     -------------------------
!     Declaration of arguments:
!     -------------------------
      implicit none

      integer n
      double precision :: A(*)
      double precision :: x(*), b(*), r(*)
      double precision :: work(n,*)
      double precision :: tolrnc(*)
      double precision :: lu(*)

      integer colind(*), rwptr(*)
      integer wksz
      integer maxits, ierr
      integer jlu(*), uptr(*)
      integer pcmeth, iunit(*), instlvl
      logical reltv

!     ------------------
!     External functions
!     ------------------
      double precision, external :: ddot,dnrm2

!     ------------- 
!     Local storage 
!     ------------- 
      logical first
      integer i, its, resets
      double precision :: alfa, beta, rho, rhonew, tau, res, tolsq
      double precision :: SMALL, pdr, urnorm

!     ------------------------------
!     Pointers to work array vectors
!     ------------------------------
      integer rh,q,v,p,ph,uh,u,t,wins1,wins2

!=======================================================================
!     Start of executable statements
!=======================================================================

!     ------------------------------
!     Set up square root of safe min
!     ------------------------------
      SMALL = 1.0d0/sqrt(huge(SMALL))

!     ---------------------------------
!     Set up inverse of machine epsilon
!     ---------------------------------
      pdr = 1.0d0/epsilon(SMALL)

!----------------------------------------------------------------------
!
!  Key to additional work storage requirements (if no instrumentation):
!         rh -> work(:,1)
!         q  -> work(:,2)
!         ph -> work(:,2)  (overlapped with q)
!         v  -> work(:,3)
!         t  -> work(:,3)  (overlapped with v)
!         p  -> work(:,4)
!         uh -> work(:,5)
!         u  -> work(:,6)
!
!----------------------------------------------------------------------

!     -------------------------------------
!     Set up pointers to work array vectors
!     -------------------------------------
      if (instlvl .gt. 0) then
         wins1 = 1
         wins2 = wins1 + 1
      else
         wins1 = 0
         wins2 = 0
      end if
         rh = wins2 + 1
         q  = rh + 1
         ph = q
         t  = ph + 1 
         v  = t
         p  = v + 1
         uh = p + 1
         u  = uh + 1

!     -------------------------------------
!     Stop if not enough workspace provided
!     -------------------------------------
      if (wksz .lt. (wins2+6)*n)  then
         ierr = wksz - (wins2+6)*n
         maxits = 0
         return
      end if

!     -------------------------
!     Initialize for iterations
!     -------------------------
      first = .true.
      tolsq = tolrnc(1)*tolrnc(1)

!     ------------
!     rh = b - A*x
!     ------------
      call bmux(n, n,x,r,A,colind,rwptr)
      do i = 1,n
         work(i,rh) = b(i) - r(i)
      end do
      urnorm = ddot(n, work(1,rh), 1, work(1,rh), 1)

!     -------------------------------------------------
!     Adjust tolerance if relative residual to be used:
!     -------------------------------------------------
      if (reltv) then
          tolsq = tolsq*urnorm
          tolrnc(1) = tolrnc(1)*sqrt(urnorm)
      end if

      if (urnorm .lt. tolsq) go to 900

!     -------------------
!     rh = r = inv(LU)*rh
!     -------------------
      call lusl(n, work(1,rh), r, lu, jlu, uptr, pcmeth,    &
                A,colind,rwptr)
      call dcopy(n,r,1,work(1,rh),1)

!     ---------
!     p = q = 0
!     ---------
      do i = 1,n
         work(i,p) = 0.d0
         work(i,q) = 0.d0
      end do

      rho = ddot(n,work(1,rh),1,r,1)

!     -------------------------------------------------------------
!     Check to see if preconditioned residual is too large relative
!     to unpreconditioned residual.
!     -------------------------------------------------------------
      if (rho .gt. pdr*pdr*urnorm) then
         ierr = 4
         maxits = 0
         return
      end if 

!     -----------------------
!     Further initializations
!     -----------------------
      beta = rho
      its = 0
      resets = 0

!=======================================================================
!     Begin main loop of iterations.  Loop is equivalent to
!      do while (rnorm .gt. tolsq) .and. (its .le. maxits)
!=======================================================================

 40      continue

         call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,  &
                     pcmeth,first,its, iunit,instlvl)
         its = its + 1

!        -------------------------
!        u = r + beta*q
!        p = u + beta*(q + beta*p)
!        -------------------------
         do i = 1,n
            work(i,u) = r(i)+beta*work(i,q)
            work(i,p) = work(i,u)+beta*(work(i,q)+beta*work(i,p))
         end do

!        ---------------
!        v = inv(LU)*A*p
!        ---------------
         call bmux(n, n,work(1,p),work(1,ph),A,colind,rwptr)
         call lusl(n, work(1,ph), work(1,v), lu, jlu, uptr, pcmeth,     &
                   A,colind,rwptr)

!        ----------
!        tau = v'rh
!        ----------
         tau = ddot(n,work(1,rh),1,work(1,v),1)
         if (abs(tau) .le. SMALL) go to 930

        alfa = rho/tau

!        ---------------
!        q = u - alpha*v
!        ---------------
         do i = 1,n
            work(i,q) = work(i,u)-alfa*work(i,v)
         end do

!        ---------
!        v = u + q
!        ---------
         do i = 1,n
            work(i,v) = work(i,u)+work(i,q)
         end do

!        --------------
!        x = x + alfa*v
!        --------------
         call daxpy(n,alfa,work(1,v),1,x,1)

!        ----------------
!        t = inv(LU)*A*v
!        ----------------
         call bmux(n, n,work(1,v),work(1,uh),A,colind,rwptr)
         call lusl(n, work(1,uh), work(1,t), lu, jlu, uptr, pcmeth,     &
                   A,colind,rwptr)      

!        --------------
!        r = r - alfa*t
!        --------------
         call daxpy(n,-alfa,work(1,t),1,r,1)

!        -------------------------
!        rhonew = rh'r;  res = r'r
!        -------------------------
         rhonew = ddot(n,work(1,rh),1,r,1)
         res = ddot(n,r,1,r,1)
         if (res.lt.tolsq) go to 910
  60     continue

     beta = rhonew/rho
         rho = rhonew
     if (abs(rho) .le. SMALL) go to 940

         if (its .ge. maxits) go to 920
         go to 40

!=======================================================================
!                  Stopping conditions
!=======================================================================

!     ====================
!     (Almost) exact guess
!     ====================
 900  continue
      ierr = 1
      maxits = 0
      return

!     =========================================
!     Possible convergence; check true residual
!     =========================================
 910  continue
      ierr = 0
      resets = resets + 1
      if (resets .gt. 8) then
         ierr = 2
         go to 915
      end if
      call bmux(n, n,x,work(1,t),A,colind,rwptr)
      do i = 1, n
         work(i,t) = b(i) - work(i,t)
      end do
      res = ddot(n,work(1,t),1,work(1,t),1)
      if (res .gt. tolrnc(1)*tolrnc(1)) then
          tolsq = tolsq*tolrnc(1)*tolrnc(1)/res
!         write(6,*)'>>>>> tolerance reset to ', sqrt(tolsq),'  <<<<<<'
          go to 60
      end if
 915  continue
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!     ================================================
!     No convergence within maximum allowed iterations
!     ================================================
 920  continue
      ierr = 3
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!     =======
!     tau = 0
!     =======
 930  continue
      ierr = 5
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!     =======
!     rho = 0
!     =======
 940  continue
      ierr = 6
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!================= End of CGS ==========================================

      end
