!======================================================================
      subroutine gmres(a,colind,rwptr,n,                    &
                       x,r,b,                               &
                       tolrnc,reltv,maxits,ierr,            &
                       work,wksz,                           &
                       lu,jlu,uptr,pcmeth,iunit,instlvl)

!-----------------------------------------------------------------------
!                       *** GMRES ***                   
!                                   
!  Standard (nonflexible) version of GMRES(m).  The grandfather of  
!  this code is one supplied by Yousef Saad while at CSRD, and many 
!  of the names of variables have been retained.            
!                                   
!  Unlike other routines in SPLIB, this uses right preconditioning. 
!  That allows keeping track of the unpreconditioned residual, without  
!  having to worry about the preconditioned residual norm giving an 
!  (too) early termination.  Nevertheless, it is possible for the   
!  preconditioner to be unstable, and the usual check is supplied for   
!  that (see the ierr = 4 condition).
!                                   
!  Also note that maxits is used as a vector here, to pass along the    
!  maximal Krylov subspace size.                    
!                                   
!-----------------------------------------------------------------------
!
!  Arguments:
!  ----------
!
!     (Data structure for A)
!  n       : Order of matrix A.
!  A       : double precision :: array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!
!     (Required vectors)
!  x       : Vector of unknowns to be solved for.  Contains initial
!            estimate of x on entry, and contains final approximate
!            on return.
!  r       : Residual vector r = b - A*x.
!  b       : Right hand side vector of system to be solved: Ax = b
!  work    : Double precision array of additional storage.  GMRES
!            requires a matrix H for the reduced approximation in
!            the Krylov subspace, as well as a few vectors. If the
!            routine instr.f is called, an additional 2 vectors of
!            length n are required.  If not enough storage is given,
!            nothing is done and an error code is returned specifying
!            the additional storage needed (in double precision words).
!            The array must have a length (see wksz) such that
!                wksz  >=   (1+im)*n   +  im   + im*(im+1)  + 3*im+1,
!                for      Krylov space + work  + Hessenberg + c,s,rs
!            Add 2*n if instrumentation is used.
!  wksz    : Length of array work, in double precision words.  
!
!     (Preconditioner data structure)
!  lu,jlu  : Preconditioner in Modified Sparse Row format containing
!            the L and U factors together. The diagonal of U (stored in
!            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
!            contains the i-th row of L (excluding the diagonal entry) 
!            followed by the i-th row of U.  The diagonal of L has
!            all ones, and so is not stored explicitly.
!  uptr    : Integer array of length n containing the pointers to        
!            the beginning of each row of U in the matrix lu,jlu. 
!  pcmeth  : Preconditioning method to use.  See splib.f documentation
!            for the interpretation of this integer.
!
!     (Iteration control)
!  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
!            passed in as a vector in case future versions
!            set pdr and SMALL.  Since Krylov methods generally
!            use the dotproduct r'*r, a local variable tolsq is used
!            in order to avoid square roots.  Note: because the
!            preconditioned residual norm || inv(M)*r || is the natural
!            one tracked by the iterative method, but users generally
!            want the true residual norm || r || as the stopping test,
!            this routine will exit the iterations when the precond
!            residual is small enough and then test the true residual.
!            If it is not sufficiently reduced, then the local tolerance
!            tolsq is reduced and iterations resume.  This is done
!            a maximum of eight times before failure is declared and
!            the routine returns.
!  reltv   : Logical variable indicating whether or not to use a
!            relative residual as the stopping test.  If this is
!            false, the stopping test is || r_k || < tolrnc(1).  If it
!            is true, the test is || r_k || < || r_0 || * tolrnc(1).
!            Note that it is possible to modify SPLIB to use other
!            stopping tests (see routine instr.f).
!  maxits  : Integer array of length two.
!            maxits(1) ---> Maximum number of iterations allowed.
!            maxits(2) ---> Dimension im of Krylov subspace to use.
!
!      (Other stuff)
!  ierr    : Error flag; see below.
!  iunit   : Integer vector of unit numbers for the instrumentation
!            routine instr.f to use.  Not referenced by gmres, just
!            passed along to instr.f
!  instlvl : Integer giving instrumentation level.  If instlvl = 0,
!            then instr.f is not called and array work() need not
!            provide the 2n double precision words used by instr.f
!
!  Some internal variables:
!  ------------------------
!
!  first   : Logical parameter that specifies to instr.f if this is
!            the first iteration, and so if new headers need to be
!            written to the Xgraph output files.
!  SMALL   : Parameter that determines when it is not safe to divide
!            by a scalar quantity.  Error condition is set if a divisor
!            is smaller than SMALL.
!  pdr     : Parameter to determine if preconditioner is no good.
!            Experience shows that if the preconditioned residual norm
!            is more than 1/eps, where eps is the machine epsilon, then
!            even if the preconditioned residual converges, the original
!            system residual does not.  If you are not comfortable with
!            this try setting pdr = 1.0d200.  
!  rs,
!  hh,c,s  : Integer pointers in the array work(*) that give the
!            starting position of the matrices and vectors used by
!            GMRES.  Note that this is much more complicated here 
!            than in the other routines of SPLIB, because of the need
!            to store a full im x im array in work().  The 
!            correspondance between GMRES algorithm arrays and the
!            work vector indexing are given by
!               hh(i,j)  --> work( (j-1)*im + i, hh)
!                  c(i)  --> work( c + i, hh)
!                  s(i)  --> work( s + i, hh)
!                 rs(i)  --> work(rs + i, hh)
!            To make things a little easier to understand, each update
!            in the work() array is accompanied by a commented out 
!            assignment in the corresponding vectors rs(), c(), etc.
!  wins1,
!  wins2   : Integer pointers in the array work(*) that give starting
!            positions of two vectors used by instrumentation routines.
!
!======================================================================
!
! Return values:                                                           
! --------------
!
!  x      : Contains an approximate solution (upon successful return).  
!
!  ierr   : Error flag with the following meanings:          
!           ierr = -m --> Not enough workspace provided; m more double
!                         precision words needed in array work().
!           ierr = 0  --> Successful return; tolerance achieved.
!           ierr = 1  --> The initial guess satisfies the stopping test.
!                         This is often a clue that rhs not set properly.
!           ierr = 2  --> Tolerance reset 8 times in attempt to
!                         get true residual small enough.
!           ierr = 3  --> convergence not achieved in maxits iterations.  
!           ierr = 4  --> Preconditioned residual is 1/eps times
!                         larger than unpreconditioned residual.
!
!  maxits : Contains the number of actual iterations performed
!           in maxits(1); the Krylov subspace size maxits(2) is
!           unchanged.
!  tolrnc : May be modified if reltv = .true.
!                                                                       
!======================================================================
!  Routines
!  --------
!
!  lusl    : Solves M z = r.
!  bmux    : Computes matrix-vector product with A.
!  instr   : Instrumentation routine for timing, etc.
!  BLAS routines: ddot, daxpy, dscal, dgemv
!
!======================================================================
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Thu Jul 21 09:14:53 EST 1994
!
!     Original version coded by Y. Saad, from SPARSKIT I 
!
!----------------------------------------------------------------------* 

!     -------------------------
!     Declaration of arguments:
!     -------------------------
      implicit none
      integer n, maxits(2), ierr
      double precision :: work(n,*), b(n), x(n), a(*), lu(*), r(*)
      integer colind(*), rwptr(n+1), jlu(*), uptr(n)
      double precision :: tolrnc(*)
      integer wksz
      logical reltv

!     ------------- 
!     Local storage 
!     ------------- 
      logical first
      double precision :: ro, t, eps1, gam
      integer im
      integer j, i, k, ii, k1, jj, i1
      integer pcmeth, its, resets, iunit(*), instlvl
      double precision :: epsmac, SMALL, pdr
      double precision :: Oetli0, Oetli1

!     -------------------------------
!     Pointers for work array vectors
!     -------------------------------
      integer hh, c, s, rs, wins1, wins2

!     ------------------
!     External functions
!     ------------------
      double precision :: ddot, dnrm2
      external ddot, dnrm2
      double precision :: Oetli
      external Oetli

!=======================================================================
!     Start of executable statements
!=======================================================================

!     ------------------------------
!     Set up square root of safe min
!     ------------------------------
      SMALL = 1.0d0/sqrt(huge(SMALL))

!     ---------------------------------
!     Set up inverse of machine epsilon
!     ---------------------------------
      epsmac = epsilon(SMALL)
      pdr = 1.0d0/epsmac

!     ----------------------------
!     Unpack Krylov subspace size.
!     ----------------------------
      im = maxits(2)

!------------------------------------------------------------------------
!     Alias workspace pointers.  First im columns of work() are for
!     the Krylov basis vectors, next im*(im+1) are for Hessenberg
!     matrix hh, then vectors c, s, and rs of lengths im, im , im+1.
!     The pointer references are then
!                hh(i,j)  --> work( (j-1)*im + i, hh)
!                   c(i)  --> work( c + i, hh)
!                   s(i)  --> work( s + i, hh)
!                  rs(i)  --> work(rs + i, hh)
!------------------------------------------------------------------------

      if (instlvl > 0) then
          wins1 = 1
          wins2 = wins1 + 1
      else
          wins1 = 0
          wins2 = 0
      end if
      hh = im + 2 + wins2
      c = im*(im+1)
      s = c + im
      rs = s + im

!     -------------------------------------
!     Stop if not enough workspace provided
!     -------------------------------------
      if (wksz .lt. wins2*n + (im*(n+im+5)+n + 1) ) then
           ierr = wksz - (wins2*n + (im*(n+im+5)+n + 1))
           maxits(1) = 0
           return
      end if

!     Initialize for iterations
!     -------------------------
      resets = 0
      its = 0
      first = .true.
      Oetli0 = Oetli(A,rwptr,colind,n, b, x, first)

!=============================================================
!             Outer Loop Starts Here 
!=============================================================

 20     continue

!       ----------------------------------------------
!       Compute initial residual vector for this cycle
!       ----------------------------------------------
        call bmux (n, n, x, work(1,wins2+1), a, colind, rwptr)
        do j = 1,n
           work(j,wins2+1) = b(j) - work(j,wins2+1)
        end do 
        ro = dnrm2(n,work(1,wins2+1),1)

        call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,   &
                   pcmeth,first,its, iunit,instlvl)

!       ---------------------------
!       Check for early termination
!       ---------------------------
        if (abs(ro) .le. SMALL) go to 999 
        t = 1.0d0/ ro
        call dscal(n,t,work(1,wins2+1),1)

!       -----------------------------------------
!       Adjust tolerance if first outer iteration
!       -----------------------------------------
        if (its .eq. 0) then
              eps1 = tolrnc(1)
              if (reltv) then
                  tolrnc(1) = tolrnc(1)*ro
                  eps1 = eps1*ro
              end if
        end if

!       --------------------------------------------------------
!       Initialize first term of RHS of Hessenberg system.  Here
!       rs is the right hand side vector for the least squares 
!       subproblem that GMRES must solve.  It is iteratively
!       updated with the Givens rotations generated below.
!       --------------------------------------------------------
!       rs(1) = ro
        work(rs+1,hh) = ro

!       -----------------------------------------
!       Size of Krylov subspace built so far is i
!       -----------------------------------------
        i = 0

!       ======================================
!       Inner loop to build up Krylov subspace
!       ======================================
 60         continue
            i = i + 1
            its = its + 1
            i1 = i + 1

!           --------------------------------------------
!           Matrix-vector product for next Krylov vector
!           begins here, and ends with call to bmux.    
!           --------------------------------------------
            call lusl (n, work(1,wins2+i), r, lu, jlu, uptr, pcmeth,    &
                         a,colind,rwptr) 

!           ------------------------------------------------------
!           Check to see if initial preconditioned residual is too
!           large relative to initial unpreconditioned residual.
!           ------------------------------------------------------
            if (its .eq. 1 ) then
                 if (ddot(n,r,1,r,1) .gt. pdr*pdr) then
                    ierr = 4
                    maxits(1) = 0
                    return
                 end if 
            end if 

!           ---------------------------
!           Matrix-vector product ends.
!           ---------------------------
            call bmux (n,n, r, work(1,wins2+i1), a, colind, rwptr) 

!           ========================================================
!           At this point, the upper Hessenberg matrix H = V'AV has
!           been factored and stored as H = Q*hh, with Q the product
!           of Givens rotations stored in c ( = cosines) c and s
!           ( = sines), and hh the (m+1) x m matrix
!
!                          x  x  x  x  x
!                          0  x  x  x  x
!             hh =         0  0  x  x  x
!                          0  0  0  x  x
!                          0  0  0  0  x
!                          0  0  0  0  0
!
!           Next, use Gram-Schmidt process to orthonormalize v(:,i+1)
!           vs. previous v's.   This adjoins another column and row 
!           to hh, as determined by the new Krylov vector V(:,i+1),
!           to get
!
!                          x  x  x  x  x n
!                          0  x  x  x  x n
!             hh =         0  0  x  x  x n
!                          0  0  0  x  x n
!                          0  0  0  0  x n
!                          0  0  0  0  0 s
!                          0  0  0  0  0 t
!           ==========================================================

            do j = 1, i
                 t = ddot(n, work(1,wins2+j),1,work(1,wins2+i1),1)
!                hh(j,i) = t
                 work( (i-1)*im+j,hh) = t
                 call daxpy(n, -t, work(1,wins2+j), 1, work(1,wins2+i1), 1)
            end do

            t = dnrm2(n,work(1,wins2+i1),1)
!           hh(i1,i) = t
            work( (i-1)*im+i1,hh) = t
        if ( abs(t) .le. SMALL) go to 100
        t = 1.0d0/t
        call dscal(n,t,work(1,wins2+i1),1)

!           -----------------------------------------------------
!           Done with Modified Gram-Schimdt and Arnoldi step.
!           Update factorization of hh; apply previous rotations.
!           The new hh still has the form given above, but with 
!           different entries n and s.
!           -----------------------------------------------------
 100        continue

!           ----------------------------------------------------
!           Apply previous transformations to i-th column of hh
!           ----------------------------------------------------
            do k = 2, i
                 k1 = k-1
!                t = hh(k1,i)
!                hh(k1,i) = c(k1)*t + s(k1)*hh(k,i)
!                hh(k,i) = -s(k1)*t + c(k1)*hh(k,i)
                 t = work( (i-1)*im+k1, hh )
                 work( (i-1)*im+k1, hh) = work(c+k1,hh)*t +         &
                          work(s+k1,hh)*work( (i-1)*im+k, hh)
                 work( (i-1)*im+k, hh) = -work(s+k1,hh)*t +         &
                          work(c+k1,hh)*work( (i-1)*im+k,hh)
            end do
!
!           gam = sqrt(hh(i,i)**2 + hh(i1,i)**2)
            gam = sqrt(work( (i-1)*im+i,hh)**2 + work( (i-1)*im+i1,hh)**2)
            if (abs(gam) .le. SMALL) gam = epsmac

!           -----------------------------------------------------
!           Determine next plane rotation to zilch out 't' in hh 
!           shown above, and apply it to rs; note that rs(j+1) is
!           still zero so the update uses only rs(i) to create
!           the new values for rs(i) and rs(i+1).  Similarly,
!           hh(i,k) is still zero for k < i (see representation
!           of H above).
!           -----------------------------------------------------
!           c(i) = hh(i,i)/gam
            work(c+i,hh) = work( (i-1)*im+i,hh)/gam
!           s(i) = hh(i1,i)/gam
            work(s+i,hh) = work ((i-1)*im+i1,hh)/gam
!           rs(i1) = -s(i)*rs(i)
            work(rs+i1,hh) = -work(s+i,hh)*work(rs+i,hh)
!           rs(i) =  c(i)*rs(i)
            work(rs+i,hh) =  work(c+i,hh)*work(rs+i,hh)
!           hh(i,i) = c(i)*hh(i,i) + s(i)*hh(i1,i)
            work( (i-1)*im+i,hh) = work(c+i,hh)*work( (i-1)*im+i,hh)    &
                             + work(s+i,hh)*work( (i-1)*im+i1,hh)

!           ---------------------------------------------------
!           Determine new residual norm, which is just the size
!           of rs(i+1)
!           ---------------------------------------------------
            ro = abs(work(rs+i1,hh))

!           ---------------------------------------------------------
!           Three convergence tests are used: first, if the Krylov 
!           subspace is of the specified maximum allowed size,
!           secondly if in middle of building up the Krylov subspace
!           the residual is small, and thirdly if the total number
!           of mat-vec multiplies (iterations) has been reached.
!           ---------------------------------------------------------
            if (i .lt. im                           &
                  .and. (ro .gt. eps1)              &
                  .and. its .le. maxits(1) )  go to 60

!          ==========================================
!          End inner loop to build up Krylov subspace
!          ==========================================

!       -----------------------------------------------------------
!       Now compute solution x. First solve upper triangular system
!       hh(1:i,1:i) * rs(1:i) = rs(1:i)
!       -----------------------------------------------------------
!       rs(i) = rs(i)/hh(i,i)

        work(rs+i,hh) = work(rs+i,hh)/work( (i-1)*im+i,hh)
        do ii = 2,i
              k = i-ii+1
              k1 = k+1
!             t = rs(k)
              t = work(rs+k,hh)
              do j = k1,i
!                t = t-hh(k,j)*rs(j)
                 t = t-work( (j-1)*im+k,hh)*work(rs+j,hh)
              end do
!             rs(k) = t/hh(k,k)
              work(rs+k,hh) = t/work( (k-1)*im+k,hh)
        end do

!       ----------------------------------------------------------
!       Form linear combination of work(*,i)'s to get solution.
!       Note that the original code for this is commented out, and
!       replaced by a call to the BLAS2 routine for mat*vec.  That
!       original code for mat-vec product was of form
!     
!                t = rs(1)
!                do k = 1, n
!                   r(k) = work(k,1)*t
!                end do
!                do j = 2, i
!                   t = rs(j)
!                   do k = 1, n
!                      r(k) = r(k)+t*work(k,j) 
!                   end do
!                end do
!       ----------------------------------------------------------
        call dgemv ('N', n, i, 1.0d0, work(1,wins2+1), n,       &
                       work(rs+1,hh), 1, 0.0d0, r, 1)

        call lusl (n, r, r, lu, jlu, uptr, pcmeth, a,colind,rwptr)

        do k=1, n
              x(k) = x(k) + r(k)
        end do

!       Restart outer loop when necessary
!       ---------------------------------
        if (ro .le. eps1) go to 990
 260    continue
        if (its .gt. maxits(1)) go to 991

!       Else compute residual vector and continue
!       -----------------------------------------
        do j = 1,i
             jj = i1-j+1
!            rs(jj-1) = -s(jj-1)*rs(jj)
             work(rs+jj-1,hh) = -work(s+jj-1,hh)*work(rs+jj,hh)
!            rs(jj) = c(jj-1)*rs(jj)
             work(rs+jj,hh) = work(c+jj-1,hh)*work(rs+jj,hh)
        end do

        do j = 1,i1
!            t = rs(j)
             t = work(rs+j,hh)
             if (j .eq. 1)  t = t-1.0d0
             call daxpy (n, t, work(1,wins2+j), 1, work(1,wins2+1), 1) 
        end do

!=============================================================
!        Outer Loop Ends Here
!=============================================================

         go to 20


!=======================================================================
!           Stopping conditions
!=======================================================================


!     (Almost) exact guess
!     =====================
 999   continue
       ierr = 1
       maxits(1) = its
       return

!     Possible convergence; check true residual
!     =========================================
 990   continue
       ierr = 0
       resets = resets + 1
       if (resets .gt. 8) then
         ierr = 2
         go to 995
       end if
       call bmux(n,n,x,r,a,colind,rwptr)
       do j = 1, n
          r(j) = r(j) - b(j)
       end do
       t = dnrm2(n,r,1)
       if (t .gt. tolrnc(1)) then
         Oetli1 = Oetli(A,rwptr,colind,n, b, x, first)
         if (Oetli1/Oetli0 .lt. max(10.0d0/pdr, tolrnc(1)*tolrnc(1)))   &
            then
            ierr = 100
            go to 995
         end if
         eps1 = eps1*tolrnc(1)/t
!        write(6,*)'>>>>> tolerance reset to ', eps1,'  <<<<<<'
         go to 260
       end if
 995   continue
       maxits(1) = its
       call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,    &
                     pcmeth,first,its, iunit,instlvl)
       return

!      No convergence within allowed iterations 
!      ======================================== 
 991   continue
       ierr = 3
       maxits(1) = its
       call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,    &
                     pcmeth,first,its, iunit,instlvl)
       return

!================= End of GMRES ====================================== 
           end
