!================================================================
      subroutine instr(A,colind,rwptr,n,            &
                       x,b,work,                    &
                       lu,jlu,uptr,pcmeth,          &
                       first,iter, iunit,instlvl)

!================================================================
!---------Instrumentation Routine for Iterative Solvers----------
!================================================================
!
!  This routine instruments the iterative solvers, by computing and
!  printing out specified measures of the error norm every prtfrq
!  iterations (see the documentation with the common block instr.h).
!  Because some of these can be expensive to compute, the time
!  spent in this routine is excluded from the timing variables.
!  This routine also opens I/O units as indicated by the specified
!  error checks.  The unit numbers must be initialized in the main
!  calling program, and put into the array iunit of common block
!  instr.h.  BEWARE: this code will create files with certain
!  names, so check the open statements before using it!
!
!  Note that the timing mechanism assumes that there is nothing
!  occuring between the call to the preconditioner computation and
!  the beginning of the iterative phase.  Change the initialization
!  of currtm if that is not the case.
!  
!  An iterative solver, if instrumentation is specified, will call 
!  this routine after every iteration.  This routine then will
!  then calculate and print out the indicated error norms, on
!  every prtfrq iteration.
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Sun Jun  5 11:07:51 EST 1994
!
!================================================================
!-------Variables------------------------------------------------
!================================================================
!
!  a,colind,rwptr,n : data structure for the sparse matrix, in CSR format
!  
!  x,b       : current solution vector and right hand side vector
!
!  work      : work array for residual computations, etc.
!              Must have length >= 2*n if instrumentation is used.
!
!  lu,jlu,uptr,pcmeth : the preconditioner, and integer indicating which
!                      preconditioner is being used.  In combined L/U
!                      MSR format.
!  
!  first     : logical variable; true if this is initial iteration.
!              This is needed for error norms that need the initial
!              residual to divide by, and to know when to write out
!              headers to the Xgraph output files.
!
!  firstmat  : logical variable; true if this is the first time any
!              solver/precond pair is used.  In this case, files are
!              opened for Xgraph output.
!
!  iter : integer, giving iteration number.
!
!  tstart     : used to record start time of this routine, so
!               that the extra computations induced by the instrumentation
!               can be excluded from the timing variables.
!
!  Axb : logical variable indicating if need to compute A*x - b
!
!  prtnow: logical variable indicating this is an iteration to print.
!
!  errnrm:  real(real8) :: error norms ---
!        errnrms(1-3): unused currently
!        errnrms(4)  :  || Ax - b ||
!        errnrms(5)  :  || inv(M) (Ax-b) ||
!        errnrms(6)  :  || Ax - b || / || r_0 ||
!        errnrms(7)  :  || inv(M) (Ax-b) || / || inv(M) r_0 ||
!        errnrms(8)  :  || Ax - b ||_inf / ( ||A||_inf ||x||_1 + ||b||_inf)
!        errnrms(9)  :  || x - xstar ||
!        errnrms(10) :  Not used.
!        errnrms(11-16): unused currently
!
!  times: vector containing timing data, for plotting convergence
!         history versus time.
!         times(1) : time spent in computing the preconditioner.
!         times(2) : time spent in iterative solver, including pc solves
!         times(3) : time spent in A*v multiplications
!                        (includes instrumentation calls to A*v)
!         times(4) : time spent in A^T*v multiplications
!                        (includes instrumentation calls to A^T*v)
!         times(5) : time spent in applying preconditioner
!                        (includes instrumentation calls to lusl)
!         times(6) : time spent in remainder of solver  (not yet done)
!         times(7) : time to be excluded because spent in instrumentation.
!         times(8) : time to be excluded because of other initializations,
!                    or because of earlier runs with the same matrix.
!
!================================================================
!----------------Declarations------------------------------------
!================================================================
!
      use kinds
      implicit none

!     Declarations:
!     =============
      integer, intent(in) :: n, colind(*), rwptr(n+1)
      integer, intent(in) :: jlu(*), uptr(n), pcmeth
      integer, intent(in) :: instlvl
      integer             :: prtfrq

      real(real8), intent(in) :: a(*), lu(*)
      real(real8), intent(in) :: x(n), b(n)
      real(real8)             :: work(n,2)
      real(real8)             :: errnrms(20)
      real(real8), save       :: res0, Mres0, Ainf, binf

      logical                 :: first
      integer, intent(in)     :: iter
      integer, intent(in)     :: iunit(16)
      logical, save           :: errchks(10)
      include '../include/heads.inc'

!     -----------------------------------------------------
!     Common block to pass along true solution vector, if
!     error norm  || x - xstar || is desired.  You must
!     change the length declaration of xstar below to match
!     whatever you have in your calling program.
!     -----------------------------------------------------
      common /truesol/ xstar
      real(real8) :: xstar(1)

!     ---------------
!     Local variables
!     ---------------
      integer :: i
      real(real8) :: tstart, currtm, mytime
      real(real8) :: rowsum
      external :: mytime
      logical :: Axb, prtnow, firstmat
      data firstmat /.true./

!     ------------------
!     External functions
!     ------------------
      real(real8), external :: dnrm2, dasum

!================================================================
!           Begin Executable Statements
!================================================================

!     ------------------------------------------
!     Just return if no instrumentation required
!     ------------------------------------------
      if (instlvl <= 0) return

!     --------------------------------
!     Get start time for this routine.
!     --------------------------------
      tstart = mytime()

!     ---------------------------------------------
!     Printing frequency; reduce this and recompile
!     if you do not want every iteration checked.
!     ---------------------------------------------
      prtfrq = 1

!     ------------------------------------------------------------
!     If first solver/precond pair, open files for norms graphical
!     output.  If in addition it is first iteration, print out 
!     identifying header for solver/precond pair.
!     ------------------------------------------------------------
      if (first) then
         times(7)   = 0.0_real8
         errchks(4) = (instlvl >= 1)
         errchks(5) = (instlvl >= 2)
         errchks(6) = (instlvl >= 3)
         errchks(7) = (instlvl >= 4)
         errchks(8) = (instlvl >= 5)
         errchks(9) = (instlvl >= 6)
         call initwr(iunit  ,first  , errchks, n      ,firstmat )
         firstmat = .false.
      end if

!     -------------------------------------------------
!     Find if this iteration is one to print out or not
!     -------------------------------------------------
      prtnow = ( mod(iter-1,prtfrq) == 0 )
      if (.not. prtnow) go to 900

!     --------------------------------------------------------
!     If first call, need to initialize infinity norms of A, b
!     --------------------------------------------------------
      if (first) then
         if (errchks(8)) then
            binf = 0.0_real8
            Ainf = 0.0_real8
            do 40  i = 1, n
               binf   = max(abs(b(i)), binf)
               rowsum = dasum(rwptr(i+1)-rwptr(i),a(rwptr(i)),1)
               Ainf   = max(rowsum, Ainf)
  40        continue
         end if 
      end if

!     ---------------------------------------------------------
!     Get current time spent executing, total time minus time
!     to be excluded for the instrumentation, minus time spent
!     in other initializations (or possibly from earlier runs
!     with the same matrix), plus time spent in computing
!     preconditioner.
!     ---------------------------------------------------------
      currtm = tstart - times(7) - times(2)     + times(1)

!     ---------------------------------------------------------
!     Need to compute A*x - b if any of errchks(4:8) is true
!     ---------------------------------------------------------
      Axb = .false.
      do i = 4, 8
         Axb = Axb .or. errchks(i)
      end do

!     ------------------------
!     Get error norm if needed
!     ------------------------
      if (errchks(9)) then
         rowsum = 0.0_real8
         do i = 1, n
            rowsum = rowsum + (xstar(i) - x(i))**2
         end do
         errnrms(9) = sqrt(rowsum)
         write(iunit(9),2000) iter, errnrms(9)
         write(iunit(16),2010) currtm, errnrms(9)
      end if

!     -----------------------------
!     Exit if no other norms needed
!     -----------------------------
      if (.not. Axb) go to 900
!
!     ----------------------------------------------------
!     Otherwise, must compute Ax - b for all error checks.
!     This uses the work vector 
!     ----------------------------------------------------
      call bmux (n, n, x, work(1,1), a, colind, rwptr)
      work(1:n,1) = work(1:n,1) - b(1:n)

!     ------------------------------
!     First compute 2-norm if needed
!     ------------------------------
      if (errchks(4) .or. errchks(6)) then
         errnrms(4) = dnrm2(n,work(1,1),1)
         if (first) res0 = errnrms(4)
         if (errchks(4)) then
            write(iunit(4),2000) iter, errnrms(4)
            write(iunit(11),2010) currtm, errnrms(4)
         end if 
         if (errchks(6)) then
            errnrms(6) = errnrms(4)/res0
            write(iunit(6),2000) iter, errnrms(6)
            write(iunit(13),2010) currtm, errnrms(6)
         end if
      end if 

!     -----------------------------------
!     Secondly compute inf-norm if needed
!     -----------------------------------
       if (errchks(8)) then
          rowsum = 0.0_real8
          do i = 1, n
             rowsum = max(rowsum, abs(work(i,1)))
          end do
          errnrms(8) = rowsum/(Ainf*dasum(n,x,1) + binf)
          write(iunit(8),2000) iter, errnrms(8)
          write(iunit(15),2010) currtm, errnrms(8)
       end if 
!
!     ------------------------------------------
!     Compute preconditioned residual if needed.
!     ------------------------------------------
       if (errchks(5) .or. errchks(7)) then
          call lusl (n, work(1,1), work(1,2), lu, jlu, uptr,    &
                     pcmeth, a, colind, rwptr)
          errnrms(5) = dnrm2(n,work(1,2),1)
          if (first) Mres0 = errnrms(5)
          if (errchks(5)) then
             write(iunit(5),2000) iter, errnrms(5)
             write(iunit(12),2010) currtm, errnrms(5)
          end if 
          if (errchks(7)) then
             errnrms(7) = errnrms(5)/Mres0
             write(iunit(7),2000) iter, errnrms(7)
             write(iunit(14),2010) currtm, errnrms(7)
          end if
       end if 
!
!     ------------------------------------
!     Exit from computing norms for output
!     ------------------------------------
!
 900   continue

!      ------------------------------------------------------
!      No longer first iteration for this solver/precond pair.
!      ------------------------------------------------------
       first = .false.

!      ---------------------------------------------
!      Exclude time spent in this routine and return
!      ---------------------------------------------
       times(7) = times(7) + mytime() - tstart
       return

 2000 format(i5,3x,e21.15)
 2010 format(1pe12.5,3x,e11.5)

       end


!========================================================================
      subroutine initwr(iunit  ,first  , errchks , n      , firstmat)
!========================================================================
!
!  initwr initializes optional files for graphs of residual or
!  error norms by writing the headers for them.  The parameter curve,
!  and ifndef for variable cadyf, are for using this routine in an
!  application package where it is called to solve several systems, but
!  using the same solver/precond pair each time.  In that case, the
!  curves will be labeled as matrix1, matrix2, etc., rather than by the
!  solver/precond names.
!
!========================================================================

      use kinds
      implicit none
      integer       :: iunit(20), n
      integer, save :: curve
      logical       :: first, errchks(*), firstmat
      include '../include/heads.inc'
!----------------------------------------------------------------------
!
!  Open additional output files if needed on first call.
!
!    iunit(4): Output unit for || Ax - b ||  vs. iter
!    iunit(5): Output unit for || inv(M) (Ax-b) || vs. iter
!    iunit(6): Output unit for || Ax - b || / || r_0 || vs. iter
!    iunit(7): Output unit for || inv(M) (Ax-b) || / || inv(M) r_0 || vs. iter
!    iunit(8): Output unit for || Ax - b ||_inf / ( ||A||_inf ||x||_1 + ||b||_inf) vs. iter
!    iunit(9): Output unit for || x - xstar || vs. iter
!    iunit(10): Not used.
!
!    iunit(11): Output unit for || Ax - b || vs. time
!    iunit(12): Output unit for || inv(M) (Ax-b) || vs. time
!    iunit(13): Output unit for || Ax - b || / || r_0 || vs. time
!    iunit(14): Output unit for || inv(M) (Ax-b) || / || inv(M) r_0 || vs. time
!    iunit(15): Output unit for || Ax - b ||_inf / ( ||A||_inf ||x||_1 + ||b||_inf) vs. time
!    iunit(16): Output unit for || x - xstar || vs. time
!
! File Names:
!
!  resi    : Output file with residual norm vs. iteration number.
!  rest    : Output file with residual norm vs. elapsed time.
!
!  reli    : Output file with relative residual norm vs. iteration number.
!  relt    : Output file with relative residual norm vs. elapsed time.
!
!  Mresi   : Output file with preconditioned residual norm vs. iteration number.
!  Mrest   : Output file with preconditioned residual norm vs. elapsed time.
!
!  Mreli   : Output file with preconditioned relative residual norm vs. iteration number.
!  Mrelt   : Output file with preconditioned relative residual norm vs. elapsed time.
!
!  oetli   : Output file with Oettli and Prager norm vs. iteration number.
!  oetlt   : Output file with Oettli and Prager norm vs. elapsed time.
!
!  errori  : Output file with error norm vs. iteration number (if lsol = true).
!  errort  : Output file with error norm vs. time (if lsol = true).
!
!  About_SPLIB.html :  A brief description of SPLIB itself.
!
!----------------------------------------------------------------------
      if (firstmat) then
         curve = 0
         if (errchks(4)) then
            open(iunit(4), file='resi')
            open(iunit(11),file='rest')
         end if
         if (errchks(5)) then
            open(iunit(5), file='Mresi')
            open(iunit(12),file='Mrest')
         end if
         if (errchks(6)) then
            open(iunit(6), file='reli')
            open(iunit(13),file='relt')
         end if
         if (errchks(7)) then
            open(iunit(7), file='Mreli')
            open(iunit(14),file='Mrelt')
         end if
         if (errchks(8)) then
            open(iunit(8), file='oetli')
            open(iunit(15),file='oetlt')
         end if
         if (errchks(9)) then
            open(iunit(9), file='errori')
            open(iunit(16),file='errort')
         end if
      end if

! ---------------------------------------------------------------
!  If printing is requested for residual norms or error norms,
!  set up header information allowing picgraph/xgraph to be used.
! ---------------------------------------------------------------
      if (.not. firstmat) go to 60
         if (errchks(4)) then
            write(iunit(4),3000) n
            write(iunit(4),3040)
            write(iunit(11),3010) n
            write(iunit(11),3040)
         end if 
         if (errchks(5)) then
            write(iunit(5),3000) n
            write(iunit(5),3050)
            write(iunit(12),3010) n
            write(iunit(12),3050)
         end if 
         if (errchks(6)) then
            write(iunit(6),3000) n
            write(iunit(6),3060)
            write(iunit(13),3010) n
            write(iunit(13),3060)
         end if 
         if (errchks(7)) then
            write(iunit(7),3000) n
            write(iunit(7),3070)
            write(iunit(14),3010) n
            write(iunit(14),3070)
         end if 
         if (errchks(8)) then
            write(iunit(8),3000) n
            write(iunit(8),3080)
            write(iunit(15),3010) n
            write(iunit(15),3080)
         end if 
         if (errchks(9)) then
            write(iunit(9),3000) n
            write(iunit(9),3090)
            write(iunit(16),3010) n
            write(iunit(16),3090)
         end if 
  60  continue

         if (errchks(4)) then
            write(iunit(4),3020) solver,precond,ipcprm
            write(iunit(11),3020) solver,precond,ipcprm
         end if 
         if (errchks(5)) then
            write(iunit(5),3020) solver,precond,ipcprm
            write(iunit(12),3020) solver,precond,ipcprm
         end if 
         if (errchks(6)) then
            write(iunit(6),3020) solver,precond,ipcprm
            write(iunit(13),3020) solver,precond,ipcprm
         end if 
         if (errchks(7)) then
            write(iunit(7),3020) solver,precond,ipcprm
            write(iunit(14),3020) solver,precond,ipcprm
         end if 
         if (errchks(8)) then
            write(iunit(8),3020) solver,precond,ipcprm
            write(iunit(15),3020) solver,precond,ipcprm
         end if 
         if (errchks(9)) then
            write(iunit(9),3020) solver,precond,ipcprm
            write(iunit(16),3020) solver,precond,ipcprm
         end if 
 
! Formats for using xgraph:

 3000  format('TitleText: Problem size N = ',i7,/,      &
           'LogY: True',/,      &
           'XUnitText: Iteration Number')
 3010  format('TitleText: Problem size N = ',i7,/,      &
           'LogY: True',/,      &
           'XUnitText: Elapsed Time')
 3020  format(/,'"',a5,'/',a4,'(',i4,')')
 3030  format(/,'"Matrix ',i3)
 3040  format('YUnitText:  Residual')
 3050  format('YUnitText:  Precond Resid')
 3060  format('YUnitText:  Rel Resid')
 3070  format('YUnitText:  Precond Rel Resid')
 3080  format('YUnitText:  Oettli Resid')
 3090  format('YUnitText:  Error')

    return
    end
